"""
Solar Radio Source

Calculate the position in horizontal coordinates for the sun.
"""
#
# Copyright (c) Tim Molteno 2013. tim@elec.ac.nz
#
import math

from tart.imaging import radio_source, tart_util
from tart.util import angle


class Sun(radio_source.RadioSource):
    """
    A RadioSource descendent that represents the sun.
    """

    def __init__(self, jy=1200.0):
        radio_source.RadioSource.__init__(self, r=149e9, jy=jy, width=0.5)

    def __repr__(self):
        return "SUN"

    """ Compute ecliptic longitude of the Sun: latitude B is assumed to be zero"""

    def solar_longitude_to_RA(self, L, utc_date):  # require L to be an angle object!!!
        jd = tart_util.JulianDay(utc_date)
        D = jd - 2451545.0  # where jd is the Julian date of interest. Then compute
        # Mean anomaly of the Sun:
        g = angle.from_dms(357.529 + 0.98560028 * D)
        # g = angle.to_rad(g)
        # Mean longitude of the Sun:
        q = 280.459 + 0.98564736 * D
        R = (
            1.00014 - 0.01671 * g.cos() - 0.00014 * math.cos(g.to_rad() * 2)
        )  # The distance of the Sun from the Earth, R, in astronomical units (AU)
        e = angle.from_dms(
            23.439 - 0.00000036 * D
        )  # mean obliquity of the ecliptic, in degrees:
        # L = angle.to_rad()
        # e = angle.to_rad(e)
        tan_RA = e.cos() * L.sin() / L.cos()
        sin_d = e.sin() * L.sin()
        # RA = math.atan(tan_RA)
        RA = angle.atan2(e.cos() * L.sin(), L.cos())
        RA = RA.to_ra()
        delta = angle.asin(sin_d)
        return RA, delta

    # http://aa.usno.navy.mil/faq/docs/SunApprox.php
    def solar_longitude(self, utc_date):
        jd = tart_util.JulianDay(utc_date)
        D = jd - 2451545.0
        # Mean anomaly of the Sun:
        g = 357.529 + 0.98560028 * D
        # Mean longitude of the Sun:
        q = 280.459 + 0.98564736 * D
        g = angle.from_dms(g)
        # Geocentric apparent ecliptic longitude of the Sun (adjusted for aberration):
        L = q + 1.915 * g.sin() + 0.020 * math.sin(g.to_rad() * 2)
        L = angle.from_dms(angle.wrap_360(L))
        return L

    # Get the RA and Declination
    def radec(self, utc_date):
        L = self.solar_longitude(utc_date)
        RA, declination = self.solar_longitude_to_RA(L, utc_date)
        return RA, declination



import numpy as np
from scipy.signal import hilbert

from tart.imaging.correlator import Correlator, corr_b, corr_b_pat
from tart.operation import observation
from tart.util import angle, utc

TEST_SCOPE_CONFIG = "tart/test/test_telescope_config.json"

import unittest


class TestCorrelator(unittest.TestCase):
    def test_correlator_simp(self):

        import numpy as np

        from tart.operation import settings
        from tart.simulation import radio, simulation_source

        N = 20
        a0 = np.random.randint(0, 2, 2 ** N)
        a1 = np.random.randint(0, 2, 2 ** N)
        self.assertAlmostEqual(corr_b(a0, a1, 2**N), corr_b_pat(a0, a1))



        t = utc.now()
        c = settings.from_file(TEST_SCOPE_CONFIG)
        c.Dict["num_antenna"] = 2
        d = [a0, a1]
        o = observation.Observation(timestamp=t, config=c, data=d)

        cor = Correlator(van_vleck_corr=True)
        vis_a = cor.correlate(o, mode="roll")
        # vis_b = cor.correlate_roll(o)
        vis_c = cor.correlate(o, mode="fftw_hilbert")
        # print(vis_a.vis(0,1), vis_b.vis(0,1), vis_c.vis(0,1))

        sample_duration = 16.02e-1
        sample_duration = 4e-1
        config = settings.from_file(TEST_SCOPE_CONFIG)
        rad = radio.Max2769B(noise_level=np.ones(config.get_num_antenna()))
        src = simulation_source.SimulationSource(
            r = 1e15,
            amplitude=1.0,
            azimuth=angle.from_dms(0.0),
            elevation=angle.from_dms(5.0),
            sample_duration=sample_duration,
        )
        fc0 = rad.int_freq

        int_sig = np.exp(-2.0j * np.pi * fc0 * rad.baseband_timebase)

        def add_noise(sig):
            return sig + np.random.normal(0.0, 0.1, len(sig))

        def sig2binary(signal):
            return np.array([1.0 if (x >= 0) else 0.0 for x in signal])

        antsig1 = src.s_baseband(rad.baseband_timebase) * int_sig

        for fraction in np.random.uniform(-np.pi / 3, np.pi / 3, 10):
            dt = ((2.0 * np.pi) / src.omega) * fraction
            print("dt", dt)
            antsig2 = (
                src.s_baseband(rad.baseband_timebase + dt)
                * int_sig
                * np.exp(1.0j * src.omega * dt)
            )

            antsig1 = add_noise(antsig1.real)
            antsig2 = add_noise(antsig2.real)

            d = [sig2binary(antsig1), sig2binary(antsig2)]
            obs = observation.Observation(timestamp=t, config=c, data=d)

            vis_a = cor.correlate(o, mode="roll")
            # vis_b = cor.correlate_roll(o)
            vis_c = cor.correlate(o, mode="fftw_hilbert")
            vis_d = cor.correlate(o, mode="fftw_hilbert_sign")

            # print(vis_a.vis(0,1),vis_b.vis(0,1))
            print(vis_a.vis(0, 1), vis_c.vis(0, 1))
            print(vis_a.vis(0, 1), vis_d.vis(0, 1))
            # print(vis_b.vis(0,1),vis_c.vis(0,1))
            # print(vis_b.vis(0,1),vis_d.vis(0,1))
            print(vis_c.vis(0, 1), vis_d.vis(0, 1))

            # self.assertAlmostEqual(vis_a.vis(0,1),vis_b.vis(0,1),4)
            self.assertAlmostEqual(vis_a.vis(0, 1), vis_c.vis(0, 1), 2)
            self.assertAlmostEqual(vis_a.vis(0, 1), vis_d.vis(0, 1), 2)
            # self.assertAlmostEqual(vis_b.vis(0,1),vis_c.vis(0,1),4)
            # self.assertAlmostEqual(vis_b.vis(0,1),vis_d.vis(0,1),4)
            self.assertAlmostEqual(vis_c.vis(0, 1), vis_d.vis(0, 1), 2)

            vis = cor.correlate(obs, mode="roll")
            cor_out = angle.from_dms(angle.wrap_360(np.angle(vis.v[0], deg=True)))
            input_angle = angle.from_dms(
                angle.wrap_360(dt * src.omega * 180.0 / np.pi)
            )  # .to_degrees()
            # print(cor_out, input_angle)
            # print(type(cor_out),  type(input_angle))
            # print(cor_out - input_angle)
            self.assertLess(np.abs((cor_out - input_angle).to_degrees()), np.degrees(np.abs(2*fraction)))

            # d = [sig2binary(antsig1),sig2binary(antsig2)]
            # obs = observation.Observation(timestamp=t, config=c, data=d)
            # vis = cor.correlate(obs,mode='fftw_hilbert')
            # cor_out = angle.wrap_360(np.angle(vis.v[0], deg=True))
            # print(cor_out, input_angle)
            # self.assertTrue(np.abs(cor_out - input_angle) < 20.)


class TestHilbert(unittest.TestCase):
    def test_hilbert(self):
        import scipy.signal

        print("hilbert")
        x = np.sin(np.arange(10000))
        h_scipy = scipy.signal.hilbert(x)
        h_pyfftw = hilbert(x)
        [self.assertAlmostEqual(i, j, 6) for i, j in zip(h_scipy, h_pyfftw)]


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from tart.operation.settings import from_file, rotate_location
from tart.util import utc


class TestSettings(unittest.TestCase):
    '''Test Settings'''
    def test_rotate_location_30(self):
        east = np.array([1., 0., 0.])
        r_e, r_n, r_u = rotate_location(30, east)
        self.assertAlmostEqual(r_e, 0.86602540378, 5)
        self.assertAlmostEqual(r_n, 0.5, 5)

    def test_rotate_location_90(self):
        east = np.array([1., 0., 0.])
        r_e, r_n, r_u = rotate_location(90, east)
        self.assertAlmostEqual(r_e, 0, 5)
        self.assertAlmostEqual(r_n, 1, 5)

    def test_load(self):
        ts = utc.now()
        config = from_file('tart/test/test_telescope_config.json')
        self.assertTrue(config.Dict is not None)



# Copyright (C) Tim Molteno 2008-2019. All rights reserved

import math

from tart.simulation.util import Util


class Ephemeris:
    GM = 3.986005e14  # earth's universal gravitational parameter m^3/s^2
    WGS84_EARTH_ROTATION_RATE = 7.2921151467e-5
    # earth rotation rate, rad/s

    def __init__(self, in_hash):
        self.a0, self.a1, self.a2, self.a3 = in_hash[
            "a"
        ]  # alpha parameters for ionospheric model
        self.b0, self.b1, self.b2, self.b3 = in_hash[
            "b"
        ]  # beta parameters for ionospheric model

        self.svprn = in_hash["svprn"]
        self.af0 = in_hash["af0"]
        self.af1 = in_hash["af1"]
        self.af2 = in_hash["af2"]

        self.toc = in_hash["toc"]  # GpsTime

        self.iode = in_hash["iode"]
        self.crs = in_hash["crs"]
        self.deltan = in_hash["deltan"]
        self.m0 = in_hash["m0"]

        self.cuc = in_hash["cuc"]
        self.ecc = in_hash["ecc"]
        self.cus = in_hash["cus"]
        self.roota = in_hash["roota"]

        self.toe = in_hash["toe"]
        self.cic = in_hash["cic"]
        self.omega_c = in_hash["omega_c"]
        self.cis = in_hash["cis"]

        self.i0 = in_hash["i0"]
        self.crc = in_hash["crc"]
        self.omega = in_hash["omega"]
        self.omegadot = in_hash["omegadot"]

        self.idot = in_hash["idot"]
        self.codes = in_hash["codes"]
        self.weekno = in_hash["weekno"]
        self.l2flag = in_hash["l2flag"]

        self.svaccur = in_hash["svaccur"]
        self.svhealth = in_hash["svhealth"]
        self.tgd = in_hash["tgd"]

        self.tom = in_hash["tom"]
        self.fit = in_hash["fit"]
        self

    def from_hash(self, in_hash):
        ret = Ephemeris()
        ret.from_hash(in_hash)
        return ret

    def to_s(self):
        return (
            f"Ephemeris: SV={self.svprn}, toe={self.toe}, toc={self.toc}, a0={self.a0}, ecc={self.ecc}, m0={self.m0}, roota={self.roota}"
        )

    def to_hash(self):
        ret = {}
        ret["a"] = [
            self.a0,
            self.a1,
            self.a2,
            self.a3,
        ]  # alpha parameters for ionospheric model
        ret["b"] = [
            self.b0,
            self.b1,
            self.b2,
            self.b3,
        ]  # beta parameters for ionospheric model

        ret["svprn"] = self.svprn
        ret["af0"] = self.af0
        ret["af1"] = self.af1
        ret["af2"] = self.af2

        ret["toc"] = self.toc  # GpsTime

        ret["iode"] = self.iode
        ret["crs"] = self.crs
        ret["deltan"] = self.deltan
        ret["m0"] = self.m0

        ret["cuc"] = self.cuc
        ret["ecc"] = self.ecc
        ret["cus"] = self.cus
        ret["roota"] = self.roota

        ret["toe"] = self.toe
        ret["cic"] = self.cic
        ret["omega_c"] = self.omega_c
        ret["cis"] = self.cis

        ret["i0"] = self.i0
        ret["crc"] = self.crc
        ret["omega"] = self.omega
        ret["omegadot"] = self.omegadot

        ret["idot"] = self.idot
        ret["codes"] = self.codes
        ret["weekno"] = self.weekno
        ret["l2flag"] = self.l2flag

        ret["svaccur"] = self.svaccur
        ret["svhealth"] = self.svhealth
        ret["tgd"] = self.tgd

        ret["tom"] = self.tom
        ret["fit"] = self.fit
        return ret

    # \brief Find the clock correction from UTC for this SV
    #         This is done ignoring leap seconds. In other words, this is modulo 1 second.
    #
    def clock_correct(self, sow):
        # more accurate calculation
        e = self.getE(self.Tsv)
        dt = Util.check_t(self.Tsv - self.toc.sow())
        dtr = (
            -4.442807e-10 * self.ecc * self.roota * math.sin(e)
        )  # relatavistic correction
        dtsv = self.af0 + self.af1 * (dt) + self.af2 * dt * dt + dtr - self.tgd
        return dtsv

    def get_tk(self, sow):
        tk = Util.check_t(
            sow - self.toe
        )  #        Time from ephemeris reference epoch (1)
        if tk < -302400.0:
            raise RuntimeError("Invalid time %f" % tk)
        if tk > 302400.0:
            raise RuntimeError("Invalid time %f" % tk)
        return tk

    # http://home-2.worldonline.nl/~samsvl/satpos.htm
    # Get the eccentricity
    def getE0(self, sow):

        x = self.m0  # kepler's equation for eccentric anomaly ek
        y = self.m0 - (x - self.ecc * math.sin(x))
        x1 = x
        x = y
        for i in range(15):
            x2 = x1
            x1 = x
            y1 = y
            y = self.m0 - (x - self.ecc * math.sin(x))
            if abs(y - y1) < 1.0e-15:
                break
            x = (x2 * y - x * y1) / (y - y1)
        ek = x  # end of det. of ecc. anomaly

        tk = self.get_tk(sow)
        n0 = math.sqrt(Ephemeris.GM / (self.roota ** 6))  #        Computed mean motion
        n = n0 + self.deltan  #        Corrected mean motion
        mk = self.m0 + n * tk  #    mean anomaly

        x = mk  #    kepler's equation for eccentric anomaly ek
        y = mk - (x - self.ecc * math.sin(x))
        x1 = x
        x = y
        for i in range(15):
            x2 = x1
            x1 = x
            y1 = y
            y = mk - (x - self.ecc * math.sin(x))
            if abs(y - y1) < 1.0e-15:
                break
            x = (x2 * y - x * y1) / (y - y1)
        ek = x

        print(f"    E0 -> {ek}")
        return ek

    def getE(self, sow):
        a = self.roota * self.roota
        tk = self.get_tk(sow)
        n0 = math.sqrt(Ephemeris.GM / (a ** 3))  #  Computed mean motion
        n = n0 + self.deltan  #  Corrected mean motion
        m = self.m0 + n * tk  #  Mean anomaly

        # test = self.getE0(sow)
        m = Util.rem2pi(m + Util.PI2)
        e = m
        for i in range(15):
            e_old = e
            e = m + self.ecc * math.sin(e_old)
            dE = Util.rem2pi(e - e_old)
            if abs(dE) < 1.0e-15:
                break

        e = Util.rem2pi(e + Util.PI2)
        return e

    def get_sv_position(self, gt):
        return self.get_location(gt.sow())

    def get_sv_position_utc(self, utc_datetime):
        gpst = gps_time.GpsTime.from_time(utc_datetime)
        return self.get_location(gpst.sow())

    def get_location(self, sow):
        a = self.roota * self.roota  # Semi major axis
        tk = self.get_tk(sow)  #     tk = sow-@toe

        e = self.getE(sow)

        v = math.atan2(
            math.sqrt(1.0 - (self.ecc ** 2)) * math.sin(e), math.cos(e) - self.ecc
        )
        phi = v + self.omega
        phi = Util.rem2pi(phi)
        phi2 = 2.0 * phi

        cosphi2 = math.cos(phi2)
        sinphi2 = math.sin(phi2)

        u = phi + self.cuc * cosphi2 + self.cus * sinphi2
        r = a * (1.0 - self.ecc * math.cos(e)) + self.crc * cosphi2 + self.crs * sinphi2
        i = self.i0 + self.idot * tk + self.cic * cosphi2 + self.cis * sinphi2
        om = (
            self.omega_c
            + (self.omegadot - Ephemeris.WGS84_EARTH_ROTATION_RATE) * tk
            - Ephemeris.WGS84_EARTH_ROTATION_RATE * self.toe
        )
        om = Util.rem2pi(om + Util.PI2)
        print(f"w_c={self.omega_c}, wdot={self.omegadot}, om={om}")
        x1 = math.cos(u) * r
        y1 = math.sin(u) * r

        x = x1 * math.cos(om) - y1 * math.cos(i) * math.sin(om)
        y = x1 * math.sin(om) + y1 * math.cos(i) * math.cos(om)
        z = y1 * math.sin(i)
        print(

                f"ephemeris.get_location({sow}) tk={tk}, {self.to_s()} -> x:{x}"

        )
        return [x, y, z]

    def get_location_new(self, sow):
        """A sanity check from a different source
        http://gnsstk.sourceforge.net/gps_8c-source.html
        https://gssc.esa.int/navipedia/index.php/GPS_and_Galileo_Satellite_Coordinates_Computation
        """
        a = self.roota * self.roota  # Semi major axis
        tk = Util.check_t(sow - self.toe)

        Mk = self.m0 + (math.sqrt(Ephemeris.GM / (a ** 3)) + self.deltan) * tk

        E_0 = Mk
        for _ in range(7):
            E_k = Mk + self.ecc * math.sin(E_0)
            E_0 = E_k

        v_k = math.atan(
            math.sqrt(1.0 - (self.ecc ** 2)) * math.sin(E_k) / math.cos(E_k) - self.ecc
        )

        phi = v_k + self.omega
        phi2 = 2.0 * phi

        cosphi2 = math.cos(phi2)
        sinphi2 = math.sin(phi2)

        u_k = phi + self.cuc * cosphi2 + self.cus * sinphi2
        r_k = (
            a * (1.0 - self.ecc * math.cos(E_k))
            + self.crc * cosphi2
            + self.crs * sinphi2
        )
        i_k = self.i0 + self.idot * tk + self.cic * cosphi2 + self.cis * sinphi2

        lambda_k = (
            self.omega_c
            + (self.omegadot - Ephemeris.WGS84_EARTH_ROTATION_RATE) * tk
            - Ephemeris.WGS84_EARTH_ROTATION_RATE * self.toe
        )

        x1 = math.cos(u_k) * r_k
        y1 = math.sin(u_k) * r_k

        x = x1 * math.cos(lambda_k) - y1 * math.cos(i_k) * math.sin(lambda_k)
        y = x1 * math.sin(lambda_k) + y1 * math.cos(i_k) * math.cos(lambda_k)
        z = y1 * math.sin(i_k)
        return [x, y, z]

    def get_velocity(self, sow):
        loc1 = np.array(self.get_location(sow - 0.5))
        loc2 = np.array(self.get_location(sow + 0.5))
        return loc1 - loc2



import unittest
import logging
import os

import numpy as np
from copy import deepcopy

from tart.util import utc
from tart.util import angle

from tart.imaging import antenna_model
from tart.imaging import radio_source
from tart.imaging import location
from tart.imaging import calibration
from tart.imaging import elaz
from tart.imaging import imaging
from tart.simulation import skymodel
from tart.simulation import antennas
from tart.simulation import radio
from tart.operation import settings


logger = logging.getLogger()

TESTCONFIG_FILENAME = os.path.join(os.path.dirname(__file__), '../../test/test_telescope_config.json')
ANT_POS_FILE = os.path.join(os.path.dirname(__file__), '../../test/test_calibrated_antenna_positions.json')


class TestImaging(unittest.TestCase):

    def setUp(self):
        self.config = settings.from_file(TESTCONFIG_FILENAME)
        self.config.load_antenna_positions(
            cal_ant_positions_file=ANT_POS_FILE
        )

    def get_clock_sources(self, timestamp):
        # ############## HOUR HAND ###########################
        #
        # The pattern rotates once every 12 hours
        #
        hour_hand = timestamp.hour*30.0 + timestamp.minute/2.0

        hour_sources = [{'el': el, 'az': -hour_hand} for el in [85, 75, 65, 55]]

        # ############## MINUTE HAND ###########################
        #
        # The pattern rotates once every 1 hour
        #
        minute_hand = timestamp.minute*6.0 + timestamp.second/10.0

        minute_sources = [{'el': el, 'az': -minute_hand} for el in [90, 80, 70, 60, 50, 40, 30]]

        return hour_sources, minute_sources

    def get_clock_vis(self, config, timestamp):

        loc = location.Location(angle.from_dms(config.get_lat()),
                                angle.from_dms(config.get_lon()),
                                config.get_alt())

        ant_pos = config.get_antenna_positions()
        num_ant = len(ant_pos)

        ANTS = [antennas.Antenna(loc, pos) for pos in ant_pos]
        ANT_MODELS = [antenna_model.GpsPatchAntenna() for i in range(num_ant)]
        RAD = radio.Max2769B(n_samples=2**12, noise_level=np.zeros(num_ant))

        hour_sources, minute_sources = self.get_clock_sources(timestamp)

        sim_sky = skymodel.Skymodel(0, location=loc, gps=0,
                                    thesun=0, known_cosmic=0)

        for m in hour_sources + minute_sources:
            src = radio_source.ArtificialSource(loc, timestamp, r=1e6,
                                                el=m['el'], az=m['az'])
            sim_sky.add_src(src)

        sources = sim_sky.gen_photons_per_src(timestamp, radio=RAD,
                                              config=self.config, n_samp=1)

        v_sim = antennas.antennas_simp_vis(
            ANTS, ANT_MODELS, sources, timestamp, self.config, RAD.noise_level
        )

        cv = calibration.CalibratedVisibility(v_sim)

        return cv, hour_sources, minute_sources

    def test_imaging(self):

        cv, hour_sources, minute_sources = \
            self.get_clock_vis(timestamp=utc.now(), config=self.config)

        n_bin = 2 ** 8

        cal_ift, cal_extent, n_fft, bin_width = imaging.image_from_calibrated_vis(
            cv, nw=n_bin / 4, num_bin=n_bin
        )

        img = np.abs(cal_ift)
        max_p = np.max(img)
        min_p = np.min(img)
        mad_p = np.std(img)
        ift_scaled = (img - min_p) / mad_p

        import matplotlib.pyplot as plt

        for m in minute_sources:
            src = elaz.ElAz(m['el'], m['az'])
            x, y = src.get_px(n_bin)
            print(x, y)
            ift_scaled[x, y] = -1
        plt.imshow(ift_scaled)

        plt.show()

        for m in minute_sources:
            src = elaz.ElAz(m['el'], m['az'])
            x, y = src.get_px(n_bin)
            self.assertGreater(img[x, y], max_p/3)



#
# An object that represents a location in the Elevation and Azimuth, and
# handles projection into the l,m plane.
#
# Azimuth is measured clockwise from north (y - axis)
# Tim Molteno 2017-2019
#
import numpy as np

from tart.util.angle import wrap_180
from tart.imaging import imaging



class ElAz:
    def __init__(self, el, az):

        self.el = el
        self.az = wrap_180(az)
        self.el_r = np.radians(self.el)
        self.az_r = np.radians(self.az)

        # l,m are the cosines of the angle between the source vector and the
        # u and v axes respectively.
        self.l = -np.sin(self.az_r) * np.cos(self.el_r)
        self.m = np.cos(self.az_r) * np.cos(self.el_r)
        self.n = np.sin(self.el_r)

    def get_lm(self):
        return self.l, self.m

    def __repr__(self):
        return f"tart.imaging.ElAz(el={self.el:5.2f}, az={self.az:5.2f})"

    '''
        Return the area of a unit-solid area of source dS
        in the l,m plane
    '''
    def get_lm_area(self, dS=1.0):
        return dS * np.sqrt(1.0 - self.l * self.l - self.m * self.m)


    def get_px(self, num_bins):
        ''' Get source location in pixels from it's direction
            cosines. These should be image coordinates in an
            image num_bins x num_bins essentially used as array
            indices.
        '''
        x_px = imaging.get_l_index(self.l, image_size = num_bins)
        y_px = imaging.get_m_index(self.m, image_size = num_bins)
        return x_px, y_px

    @classmethod
    def from_pixels(cls, x_pix, y_pix, num_bins):
        ''' Create from pixel coordinates. The center (0,0) is
            in the middle.
        '''
        # print(f"from_pixels({x_pix}, {y_pix})")
        n2 = num_bins // 2
        x = x_pix - n2
        y = num_bins - y_pix - n2    # y_px = num_bins - int(np.round(self.m * n2 + n2))
                                     # num_bins - y_px - n2 =  mn2
                                     # num_bins - y_px - n2 =  mn2
        # print(f"centred_pixels({x}, {y})")
        r = np.sqrt(x*x + y*y)
        # print(f"r=({r}, {n2})")

        if r < n2:
            el_r = np.arccos(r/n2)
        else:
            el_r = 0.0
        # print(f"    el_r = {el_r}")
        atn = np.arctan2(y, x)
        # print(f"    atn = {np.degrees(atn)}")
        az_r = atn - np.radians(90)
        # print(f"    az_r = {az_r}")

        return ElAz(np.degrees(el_r), np.degrees(az_r))

    def get_px_window(self, num_bins, window_deg):
        ''' Get a pixel window around a source with width window_deg
            This is assumed to be an inverse FFT image with num_bins x num_bins
        '''
        x_i, y_i = self.get_px(num_bins)

        d = imaging.deg_to_pix(num_bins, window_deg)

        x_min = int(np.floor(x_i - d))
        x_max = int(np.ceil(x_i + d))
        y_min = int(np.floor(y_i - d))
        y_max = int(np.ceil(y_i + d))
        area = (x_max - x_min) * (y_max - y_min)
        return x_min, x_max, y_min, y_max, area

    def get_old_lm(self):
        R = 10 * np.pi * 2
        r = 90 - np.degrees(self.el_r)
        r = R * np.arctan(r / R)
        x = r * np.sin(self.az_r)
        y = r * np.cos(self.az_r)
        return [-x, y]


def from_json(source_json, el_limit=0.0, jy_limit=1e5):
    src_list = []

    for src in source_json:
        try:
            if (src["el"] > el_limit) and (src["jy"] > jy_limit):
                src_list.append(ElAz(src["el"], src["az"]))
        except:
            print(f"ERROR in catalog src={src}")

    return src_list


def get_source_coordinates(source_list):
    x_list = []
    y_list = []

    for src in source_list:
        x_list.append(src.l)
        y_list.append(src.m)

    return [x_list, y_list]



import unittest

import numpy as np

from tart.imaging import antenna_model
from tart.util import angle, utc


def create_empirical_antenna():
    sv = 1
    ant = antenna_model.EmpiricalAntenna(1)
    for e in np.arange(0.0, 89.0, 2.0):
        for a in np.arange(0.0, 359.0, 5.0):
            el = angle.from_dms(e)
            az = angle.from_dms(a)
            ant.add_measurement(el, az, e * (1 + np.cos(np.pi / 180.0 * a)), sv)
    for e in np.arange(0.0, -89.0, -2.0):
        for a in np.arange(0.0, 359.0, 5.0):
            el = angle.from_dms(e)
            az = angle.from_dms(a)
            ant.add_measurement(el, az, 0.0, sv)
    return ant


class TestAntennaModel(unittest.TestCase):
    """Test our antenna models."""

    def test_gps(self):
        """Test the GPS patch antenna."""
        ant = antenna_model.GpsPatchAntenna()
        self.assertEqual(ant.get_gain(angle.from_dms(120), angle.from_dms(120)), 0.0)
        self.assertEqual(ant.get_gain(angle.from_dms(80), angle.from_dms(120)), 1.0)
        self.assertEqual(ant.get_gain(angle.from_dms(-80), angle.from_dms(120)), 0.0)

    def test_ideal(self):
        """Test the ideal hemispherical antenna"""
        ant = antenna_model.IdealHemisphericalAntenna()
        self.assertEqual(ant.get_gain(angle.from_dms(120), angle.from_dms(120)), 0.0)
        self.assertEqual(ant.get_gain(angle.from_dms(80), angle.from_dms(120)), 1.0)
        self.assertEqual(ant.get_gain(angle.from_dms(-80), angle.from_dms(120)), 0.0)

    # def test_empirical(self):
    #'''Test the empirical hemispherical antenna'''
    # ant = create_empirical_antenna()
    ##self.assertAlmostEqual(ant.get_gain(angle.from_dms(70), angle.from_dms(120)) / (70.*(1+np.cos(np.pi/180.*120.))), 1.0 , 1)
    ##self.assertAlmostEqual(ant.get_gain(angle.from_dms(45), angle.from_dms(45)) / (45.*(1+np.cos(np.pi/180.*45.))), 1.0, 1)
    ##self.assertAlmostEqual(ant.get_gain(angle.from_dms(-80), angle.from_dms(120)), 0.0, 1)
    # for e in np.arange(0., 89., 12.0):
    # for a in np.arange(0., 359., 25.0):
    # el = angle.from_dms(e)
    # az = angle.from_dms(a)
    # if e != 0:
    # check = ant.get_gain(el, az) - e*(1.+np.cos(np.pi/180.*a))
    # self.assertAlmostEqual(check, 0 , 1)

    def test_json_load(self):
        ant = create_empirical_antenna()
        ant.to_json("test.json")
        ant2 = antenna_model.EmpiricalAntenna.from_json("test.json")
        for e in np.arange(0, 89, 3.0):
            for a in np.arange(0, 359, 8.0):
                el = angle.from_dms(e)
                az = angle.from_dms(a)
                self.assertAlmostEqual(ant.get_gain(el, az), ant2.get_gain(el, az), 3)

    def test_db_load(self):
        ant = create_empirical_antenna()
        ant.to_db(utc_date=utc.now(), db_file="test.db")
        ant2 = antenna_model.EmpiricalAntenna.from_db(antenna_num=ant.antenna_num, db_file="test.db")
        for e in np.arange(0, 89, 3.0):
            for a in np.arange(0, 359, 8.0):
                el = angle.from_dms(e)
                az = angle.from_dms(a)
                self.assertAlmostEqual(ant.get_gain(el, az), ant2.get_gain(el, az), 3)





import unittest

from tart.imaging import location
from tart.simulation import skymodel


class TestSkymodel(unittest.TestCase):
    """Testharness for Skymodel class"""

    def test_from_to_state_vector(self):
        """Generate Model from state vector. Then generate state vector from model and compare"""
        sm = skymodel.Skymodel(5, location.Dunedin)
        state = sm.get_state_vector()
        sm2 = skymodel.from_state_vector(state)
        state2 = sm2.get_state_vector()
        print(state)
        print(state2)

        # sv_init = np.array([15, 0, 90, -60, 2, 2, 0.5, 0.1])
        # model = from_state_vector(sv_init)
        # sv_final = model.get_state_vector()
        # for i, f in zip(sv_init, sv_final):
        #   self.assertAlmostEqual(i, f, 10)



"""
RadioSource Radio Source

Some bright cosmic radio sources
"""
#
# Copyright (c) Tim Molteno 2013. tim@elec.ac.nz
#

from tart.util import angle, skyloc


class RadioSource:
    """This is the base class for all possible radio sources"""

    def __init__(self, r, jy=1000.0, width=0.01):
        self.flux = jy
        self.width = width
        self.r = r

    def radec(self, utc_date):  # Get the RA and Declination
        raise RuntimeError("Use a class derived from RadioSource")

    def jansky(self, utc_date):
        """ Spectral flux density S. This is the spectral radiance B(theta, phi) integrated over the solid angle of the source. """
        return self.flux

    def to_horizontal(self, location, utc_date):
        RA, declination = self.radec(utc_date)
        el, az = location.equatorial_to_horizontal(utc_date, RA, declination)
        return el, az


class ArtificialSource(RadioSource):
    def __init__(self, location, utc_time, r, el, az, jy=1000.0, width=0.001):
        RadioSource.__init__(self, r, jy, width)
        a_az = angle.from_dms(az)
        a_el = angle.from_dms(el)
        self.skyloc = skyloc.Skyloc.from_horizontal(location, utc_time, a_el, a_az)

    def radec(self, utc_date):  # Get the RA and Declination
        return self.skyloc.ra, self.skyloc.dec


class CosmicSource(RadioSource):
    """
    A cosmic radio source at fixed equatorial coordinates.
    """

    def __init__(self, ra, dec, r, jy=1000.0, width=0.001):
        RadioSource.__init__(self, r, jy, width)
        self.skyloc = skyloc.Skyloc(ra, dec)

    def __repr__(self):
        return "COSMIC" + str(self.skyloc)

    def radec(self, utc_date):  # Get the RA and Declination
        return self.skyloc.ra, self.skyloc.dec



# Interpolator for sp3 orbit files
# Copyright 2013 (c) Tim Molteno. tim@elec.ac.nz


class Sp3Interpolator:
    def __init__(self, gpst, points):
        self.interps = points
        self.week = gpst.week()

    def lagrange_numerator(self, t_i, i, t):
        n = len(t_i)
        ret = 1.0
        for j in range(n):
            if j != i:
                ret *= t - t_i[j]
        return ret

    def lagrange_denominator(self, t_i, i):
        n = len(t_i)
        ret = 1.0
        for j in range(n):
            if j != i:
                ret *= t_i[i] - t_i[j]
        return ret

    def lagrange_interpolate(self, t_i, y_i, t):
        ret = 0.0
        n = len(t_i)
        for i in range(n):
            ret += y_i[i] * (
                self.lagrange_numerator(t_i, i, t) / self.lagrange_denominator(t_i, i)
            )
        return ret

    def get_sv_position(self, gt, sv):
        sow = gt.sow()

        inter = self.interps[sv]

        dt0 = 999999.0
        i_close = 0
        n = len(inter)
        for j in range(n):
            i = inter[j]
            test = abs(i[0] - sow)
            if (i[0] <= sow) and (test < dt0):
                dt0 = test
                i_close = j

        t_points = []
        x_points = []
        y_points = []
        z_points = []

        order = 5
        i0 = max(0, i_close - order)
        i1 = min(i_close + order + 1, n - 1)

        for i in range(i0, i1):
            t_points.append(inter[i][0])
            x_points.append(inter[i][1])
            y_points.append(inter[i][2])
            z_points.append(inter[i][3])

        xl = self.lagrange_interpolate(t_points, x_points, sow) * 1000.0
        yl = self.lagrange_interpolate(t_points, y_points, sow) * 1000.0
        zl = self.lagrange_interpolate(t_points, z_points, sow) * 1000.0

        return [xl, yl, zl]



#
# A class the encapsulates a CalibratedVisibility object.
#
# Copyright (c) Tim Molteno 2013-2020. tim@elec.ac.nz
#

import json

import numpy as np


class CalibratedVisibility:
    def __init__(self, vis):
        self.vis = vis
        self.flagged_baselines = []
        self.phase_offset = np.zeros(self.get_config().get_num_antenna())
        self.tile_phase_offset = np.zeros(4)
        self.gain = np.ones(self.get_config().get_num_antenna())

    def set_config(self, config):
        self.vis.config = config

    def get_config(self):
        return self.vis.config

    def get_timestamp(self):
        return self.vis.timestamp

    def __get_vis(self, bl):
        return self.vis.v[self.vis.baselines.index(bl)]

    def flag_baseline(self, i, j):
        bl = [i, j]
        if bl not in self.flagged_baselines:
            # vis_idx = self.vis.baselines.index(bl)
            self.flagged_baselines.append(bl)
            # print('flagged', vis_idx, bl)

    def flag_antenna(self, i):
        for bl in self.vis.baselines:
            if i in bl:
                [a, b] = bl
                self.flag_baseline(a, b)

    def get_visibility(self, i, j):
        if j < i:
            return np.conjugate(self.get_visibility(j, i))

        bl = [i, j]
        if bl in self.flagged_baselines:
            raise
        else:
            v = self.__get_vis(bl)
            return (
                v
                * self.get_gain(i)
                * self.get_gain(j)
                * np.exp(-1j * (self.get_phase_offset(i) - self.get_phase_offset(j)))
            )

    def get_all_visibility(self):
        bl = np.asarray(self.get_baselines())
        v_arr = np.asarray(self.get_unflagged_vis())
        ret = (
            v_arr
            * self.get_gain(bl[:, 0])
            * self.get_gain(bl[:, 1])
            * np.exp(
                -1j
                * (self.get_phase_offset(bl[:, 0]) - self.get_phase_offset(bl[:, 1]))
            )
        )
        return ret, bl

    def get_all_uvw(self):
        c = self.get_config()
        ant_p = np.asarray(c.get_antenna_positions())

        bls = np.asarray(self.get_baselines())
        bl_pos = ant_p[bls]
        uu_a, vv_a, ww_a = (bl_pos[:, 0] - bl_pos[:, 1]).T
        return uu_a, vv_a, ww_a

    def get_unflagged_vis(self):
        return [
            vis
            for bl, vis in zip(self.vis.baselines, self.vis.v)
            if bl not in self.flagged_baselines
        ]

    def get_baselines(self):
        return [bl for bl in self.vis.baselines if bl not in self.flagged_baselines]

    def get_baseline_lengths(self):
        pos = self.get_config().get_antenna_positions()
        # print(pos)
        ret = []
        for bln in self.get_baselines():
            [i, j] = bln
            ret.append([pos[j][0] - pos[i][0], pos[j][1] - pos[i][1]])
        return np.abs(ret)

    def set_flagged_baselines(self, flaged_list):
        self.flagged_baselines = flaged_list

    def flag_tile(self, tile_idx):
        for bln in self.get_baselines():
            [i, j] = bln
            if ((i // 6) == tile_idx) or ((j // 6) == tile_idx):
                self.flag_baseline(i, j)

    def leave_parallel_baselines(self, ew_threshold=10, ns_threshold=10):
        ant_positions = np.array(self.get_config().get_antenna_positions())
        for bln in self.get_baselines():
            [i, j] = bln
            diff = ant_positions[i] - ant_positions[j]
            if np.abs(diff[0]) > ew_threshold or np.abs(diff[1]) > ns_threshold:
                self.flag_baseline(i, j)

    def leave_intra_tile_baselines(self):
        ant_positions = np.array(self.get_config().get_antenna_positions())
        for bln in self.get_baselines():
            [i, j] = bln
            if (i // 6) != (j // 6):
                self.flag_baseline(i, j)

    def set_phase_offset(self, i, val):
        self.phase_offset[i] = np.asarray(val) % (2 * np.pi)

    def get_phase_offset(self, i):
        return self.phase_offset[i] + self.get_tile_phase_offset_for_antenna(i)

    def set_tile_phase_offset(self, i, val):
        self.tile_phase_offset[i] = val

    def get_tile_offset(self, i):
        return self.tile_phase_offset[i]

    def get_tile_phase_offset_for_antenna(self, ant_idx):
        return self.tile_phase_offset[np.asarray(ant_idx) // 6]

    def set_gain(self, i, val):
        self.gain[i] = val

    def get_gain(self, i):
        return self.gain[i]

    def to_json(self, filename="gain_calibration.json"):
        calib_dict = {}
        calib_dict["gain"] = [
            int(1e5 * self.get_gain(i)) / 1e5 for i in range(len(self.gain))
        ]
        calib_dict["phase_offset"] = [
            int(1e5 * self.get_phase_offset(i)) / 1e5
            for i in range(len(self.gain))
        ]
        calib_dict["flagged_baselines"] = self.flagged_baselines
        with open(filename, "w") as handle:
            json.dump(calib_dict, handle, sort_keys=True, indent=4, ensure_ascii=False)
        json_str = json.dumps(calib_dict)
        return json_str


def from_dict(vis, calib_dict):
    ret = CalibratedVisibility(vis)
    ret.set_phase_offset(
        list(range(ret.get_config().get_num_antenna())),
        np.array(calib_dict["phase_offset"]),
    )
    ret.set_gain(
        list(range(ret.get_config().get_num_antenna())), np.array(calib_dict["gain"])
    )
    ret.set_flagged_baselines(calib_dict["flagged_baselines"])
    return ret


def from_JSON(vis, json_str):
    calib_dict = json.loads(json_str)
    return from_dict(vis, calib_dict)


def from_JSON_file(vis, filename="gain_calibration.json"):
    with open(filename) as data_file:
        calib_dict = json.load(data_file)
    return from_dict(vis, calib_dict)


import argparse

from tart.imaging import visibility

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generate visibilities for a skymodel in parallel."
    )
    parser.add_argument(
        "--vis",
        default="/freenas2/tart/visibility/vis_2015_08_10.vis",
        required=False,
        help="",
    )
    args = parser.parse_args()

    vis = visibility.list_load(args.vis)  # [:20]

    cal_vis = CalibratedVisibility(vis[0])
    # print(cal_vis.get_visibility(0,1))
    # print(cal_vis.get_baselines())
    # print(len(cal_vis.get_baselines()))

    cal_vis.flag_baseline(0, 1)

    # print(cal_vis.get_baselines())
    # print(len(cal_vis.get_baselines()))

    def gen_tile_vis_list(tile_no, vis):
        for v in vis:
            cal_vis_list = []
            cv = calibration.CalibratedVisibility(v)
            for i in [0, 1, 2, 3]:
                if i != tile_no:
                    cv.flag_tile(i)

        # cv.flag_tile(1)
        # cv.flag_tile(2)
        # cv.flag_tile(3)
        # cv.leave_parallel_baselines(ns_threshold=0.1)
        # cv.flag_baseline(0,5)
        cal_vis_list.append(cv)
        return cal_vis_list

    # t0_vis_list = gen_tile_vis_list(0,vis)
    # t1_vis_list = gen_tile_vis_list(1,vis)
    # t2_vis_list = gen_tile_vis_list(2,vis)
    # t3_vis_list = gen_tile_vis_list(3,vis)
    # fig, axarr = plt.subplots(2,2,sharex=True,sharey=True)
    # t_list = [t0_vis_list, t1_vis_list, t2_vis_list, t3_vis_list]
    # ift_arr = []
    # for i, ax in enumerate(axarr.ravel()):
    #   syn_tile = synthesis.Synthesis_Imaging(t_list[i])
    #   ift = syn_tile.get_image(nw=50,num_bin=2**9,ex_ax=ax)
    #   ift_arr.append(ift)

    # ift_arr = np.array(ift_arr)
    # print(ift_arr)
    # plt.figure()
    # plt.imshow(np.prod(np.abs(ift_arr),axis=0))

    # print(cal_vis.get_visibility(0,1))



import os

import numpy as np

from tart.imaging import ephemeris, gps_time, sp3_interpolator
from tart.util.singleton import Singleton


@Singleton
class EphemeridesProxy:
    def __init__(self):

        if "EPHEMERIS_SERVER_HOST" in os.environ:
            server_host = os.environ["EPHEMERIS_SERVER_HOST"]
        else:
            server_host = "localhost"

        import jsonrpclib
        self.server = jsonrpclib.Server("http://%s:8876/rpc/gps" % server_host)
        self.cache = {}
        self.sp3_cache = {}

    def get_date_hash(self, utc_date, sv):
        # Make a hash
        gpst = gps_time.GpsTime.from_time(utc_date)
        cache_hash = "%02d%04d%02d%02d%02d-%4d" % (
            sv,
            utc_date.year,
            utc_date.month,
            utc_date.day,
            utc_date.hour,
            gpst.m_week,
        )
        return cache_hash

    def get_sp3_hash(self, utc_date):
        # Make a hash
        gpst = gps_time.GpsTime.from_time(utc_date)
        cache_hash = "%04d%02d%02d-%4d" % (
            utc_date.year,
            utc_date.month,
            utc_date.day,
            gpst.m_week,
        )
        return cache_hash

    def get_ephemeris(self, utc_date, sv):
        h = self.get_date_hash(utc_date, sv)
        print(f"hash({h}")
        try:
            eph = self.cache[h]
        except KeyError:
            eph_hash = self.server.get_ephemeris(utc_date.isoformat(), sv)
            eph = ephemeris.Ephemeris(eph_hash)
            self.cache[h] = eph
            print("Cache miss %s, %d" % (utc_date, sv))
        return eph

    def get_sp3_interpolator(self, utc_date):
        h = self.get_sp3_hash(utc_date)
        try:
            sp3 = self.sp3_cache[h]
        except KeyError:
            gpst = gps_time.GpsTime.from_time(utc_date)
            pts = self.server.get_interp_points(utc_date.isoformat())
            sp3 = sp3_interpolator.Sp3Interpolator(gpst, pts)
            self.sp3_cache[h] = sp3
            print("sp3 Cache miss %s" % (utc_date))
        return sp3

    def get_sv_position(self, utc_date, sv):
        gpst = gps_time.GpsTime.from_time(utc_date)
        eph = self.get_ephemeris(utc_date, sv)
        pos = eph.get_sv_position(gpst)
        print(f"get_sv_position({utc_date}, {gpst}, {sv}) -> {pos}")
        return np.array(pos)

    def get_sv_position_sp3(self, utc_date, sv):
        gpst = gps_time.GpsTime.from_time(utc_date)
        sp3 = self.get_sp3_interpolator(utc_date)
        pos = sp3.get_sv_position(gpst, sv)
        return np.array(pos)

    def get_sv_velocity(self, utc_date, sv):
        gpst = gps_time.GpsTime.from_time(utc_date)
        eph = self.get_ephemeris(utc_date, sv)
        return eph.get_velocity(gpst.sow())

    def get_sv_positions(self, utc_date):
        gpst = gps_time.GpsTime.from_time(utc_date)
        ret = []
        for sv in range(1, 32):
            try:
                eph = self.get_ephemeris(utc_date, sv)
                pos = eph.get_sv_position(gpst)
                ret.append([sv, pos])
            except jsonrpclib.ProtocolError:
                pass
        return ret

    def get_remote_position(self, utc_date, sv):
        return self.server.get_sv_position_sp3(utc_date.isoformat(), sv)



"""
Solar Radio Source

Calculate the position in horizontal coordinates for the sun.
"""
#
# Copyright (c) Tim Molteno 2013. tim@elec.ac.nz
#
import datetime
import unittest

from tart.imaging import location, sun
from tart.util import angle, utc


class TestSun(unittest.TestCase):
    def setUp(self):
        self.t = utc.utc_datetime(2002, 10, 31, 0, 2, 2)
        self.s = sun.Sun()

    def test_horizontal(self):
        el, az = self.s.to_horizontal(location.Dunedin, self.t)
        print("El %f    Az %f" % (el.to_degrees(), az.to_degrees()))
        self.assertAlmostEqual(el.to_degrees(), 57.848294, 3)
        self.assertAlmostEqual(az.to_degrees(), 8.943156, 1)

    def test_now_horizontal(self):
        t = utc.now()
        s = sun.Sun()
        el, az = s.to_horizontal(location.Dunedin, t)
        print("El %f    Az %f" % (el.to_degrees(), az.to_degrees()))

    def test_radec(self):
        Dunedin = location.Dunedin
        import ephem

        dnd = ephem.Observer()
        dnd.lon = str(Dunedin.lon.to_degrees())
        dnd.lat = str(Dunedin.lat.to_degrees())
        dnd.elevation = Dunedin.alt
        dnd.pressure = 0.0
        datee = utc.now()
        for delta in range(24):
            timedelta = datetime.timedelta(hours=delta)
            dnd.date = datee + timedelta

            j = ephem.Sun()
            j.compute(dnd)
            ra = angle.from_rad(j.ra.real)
            dec = angle.from_rad(j.dec.real)
            s = sun.Sun()
            sra, sdec = s.radec(datee + timedelta)
            self.assertAlmostEqual(sra.to_degrees(), ra.to_degrees(), 2)
            self.assertAlmostEqual(sdec.to_degrees(), dec.to_degrees(), 2)



# Code reproduced from
# http://www.scipy.org/Cookbook/ButterworthBandpass

try:
    from scipy.signal import butter, lfilter

    SCIPY_AVAILABLE = True
except ImportError:
    SCIPY_AVAILABLE = False


def butter_bandpass(lowcut, highcut, fs, order=5):
    if not SCIPY_AVAILABLE:
        raise ImportError("scipy is required for butter_bandpass. Install with: pip install scipy")
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype="band")
    return b, a


def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
    if not SCIPY_AVAILABLE:
        raise ImportError(
            "scipy is required for butter_bandpass_filter. Install with: pip install scipy"
        )
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = lfilter(b, a, data)
    return y


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np
    from scipy.signal import freqz

    # Sample rate and desired cutoff frequencies (in Hz).
    fs = 5000.0
    lowcut = 500.0
    highcut = 1250.0

    # Plot the frequency response for a few different orders.
    plt.figure(1)
    plt.clf()
    for order in [3, 6, 9]:
        b, a = butter_bandpass(lowcut, highcut, fs, order=order)
        w, h = freqz(b, a, worN=2000)
        plt.plot((fs * 0.5 / np.pi) * w, abs(h), label="order = %d" % order)

    plt.plot([0, 0.5 * fs], [np.sqrt(0.5), np.sqrt(0.5)], "--", label="sqrt(0.5)")
    plt.xlabel("Frequency (Hz)")
    plt.ylabel("Gain")
    plt.grid(True)
    plt.legend(loc="best")

    # Filter a noisy signal.
    T = 0.05
    nsamples = T * fs
    t = np.linspace(0, T, nsamples, endpoint=False)
    a = 0.02
    f0 = 600.0
    x = 0.1 * np.sin(2 * np.pi * 1.2 * np.sqrt(t))
    x += 0.01 * np.cos(2 * np.pi * 312 * t + 0.1)
    x += a * np.cos(2 * np.pi * f0 * t + 0.11)
    x += 0.03 * np.cos(2 * np.pi * 2000 * t)
    plt.figure(2)
    plt.clf()
    plt.plot(t, x, label="Noisy signal")

    y = butter_bandpass_filter(x, lowcut, highcut, fs, order=6)
    plt.plot(t, y, label="Filtered signal (%g Hz)" % f0)
    plt.xlabel("time (seconds)")
    plt.hlines([-a, a], 0, T, linestyles="--")
    plt.grid(True)
    plt.axis("tight")
    plt.legend(loc="upper left")

    plt.show()



import unittest

import numpy as np

from tart.imaging import antenna_model, location
from tart.simulation.antennas import (
    Antenna,
    antennas_signal,
    get_geo_delay_horizontal,
    get_UVW,
)
from tart.simulation.simulation_source import HorizontalSource, SimulationSource
from tart.util import angle, constants, utc


def get_src(el, az):
    return HorizontalSource(
        r=1e9, azimuth=angle.from_dms(az), elevation=angle.from_dms(el)
    )


class TestAntenna(unittest.TestCase):
    def test_90(self):
        a0 = Antenna(location.Dunedin, [0.0, 0.0, 0.0])
        a1 = Antenna(location.Dunedin, [0.0, 1.0, 0.0])
        src = get_src(90, 0)

        self.assertAlmostEqual(a0.get_geo_delay_horizontal(src), 0.0)
        self.assertAlmostEqual(a1.get_geo_delay_horizontal(src), 0.0)
        self.assertAlmostEqual(get_geo_delay_horizontal(a0, a1, src), 0.0)

    def test_60(self):
        a0 = Antenna(location.Dunedin, [0.0, 0.0, 0.0])
        a1 = Antenna(location.Dunedin, [0.0, 1.0, 0.0])
        src = get_src(60, 0)

        self.assertAlmostEqual(a0.get_geo_delay_horizontal(src), 0.0)
        self.assertAlmostEqual(
            a1.get_geo_delay_horizontal(src) * constants.V_LIGHT, -0.5, places=4
        )  # a1 is closer to origin so negative
        self.assertAlmostEqual(
            get_geo_delay_horizontal(a0, a1, src) * constants.V_LIGHT, -0.5, places=4
        )  # t_a1 - t_a0 negative
        self.assertAlmostEqual(
            get_geo_delay_horizontal(a1, a0, src) * constants.V_LIGHT, 0.5, places=4
        )

    def test_symmetry(self):
        a0 = Antenna(location.Dunedin, [0.0, 0.0, 0.0])
        a1 = Antenna(location.Dunedin, [0.0, 1.0, 0.0])
        src = get_src(60, 180)
        self.assertAlmostEqual(a0.get_geo_delay_horizontal(src), 0.0)
        self.assertAlmostEqual(
            a1.get_geo_delay_horizontal(src) * constants.V_LIGHT, 0.5, places=4
        )  # a1 is further away, so positive

    def test_uvw(self):
        a0 = Antenna(location.Dunedin, [0.0, 0.0, 0.0])
        a1 = Antenna(location.Dunedin, [0.0, 1.0, 0.0])
        el = angle.from_dms(90.0)
        for az_deg in range(0, 360, 10):
            az = angle.from_dms(az_deg)
            utc_time = utc.now()
            ra, dec = location.Dunedin.horizontal_to_equatorial(utc_time, el, az)
            u, v, w = get_UVW(a0, a1, utc_time, ra, dec)
            self.assertAlmostEqual(u, 0.0)
            self.assertAlmostEqual(v, -1.0)
            self.assertAlmostEqual(w, 0.0)

    def test_uvw_60(self):
        a0 = Antenna(location.Dunedin, [0.0, 0.0, 0.0])
        a1 = Antenna(location.Dunedin, [0.0, 1.0, 0.0])
        el = angle.from_dms(30.0)
        az = angle.from_dms(0.0)
        utc_time = utc.now()
        ra, dec = location.Dunedin.horizontal_to_equatorial(utc_time, el, az)
        u, v, w = get_UVW(a0, a1, utc_time, ra, dec)
        self.assertAlmostEqual(u, 0.0)
        self.assertAlmostEqual(v, -0.5)
        self.assertAlmostEqual(w, -0.86602540378443871)

    def test_uvw_90(self):
        a0 = Antenna(location.Dunedin, [0.0, 0.0, 0.0])
        a1 = Antenna(location.Dunedin, [0.0, 1.0, 0.0])
        el = angle.from_dms(0.0)
        az = angle.from_dms(0.0)
        utc_time = utc.now()
        ra, dec = location.Dunedin.horizontal_to_equatorial(utc_time, el, az)
        u, v, w = get_UVW(a0, a1, utc_time, ra, dec)
        self.assertAlmostEqual(u, 0.0)
        self.assertAlmostEqual(v, 0.0)
        self.assertAlmostEqual(w, -1.0)

    def test_uvw_east(self):
        a0 = Antenna(location.Dunedin, [0.0, 0.0, 0.0])
        a1 = Antenna(location.Dunedin, [1.0, 0.0, 0.0])
        el = angle.from_dms(0.0)
        az = angle.from_dms(90.0)  # east
        utc_time = utc.now()
        ra, dec = location.Dunedin.horizontal_to_equatorial(utc_time, el, az)
        uu, vv, ww = get_UVW(a0, a1, utc_time, ra, dec)
        self.assertAlmostEqual(uu, 0.0)
        self.assertAlmostEqual(vv, 0.0)
        self.assertAlmostEqual(ww, -1.0)


class TestAntennas(unittest.TestCase):
    def test_antennas(self):
        """Test the construction of antenna signals"""
        ref_freq = 16.368e6
        freq_mult = 256
        sample_duration = 1e-5
        antenna_locations = [[-10.0, 0.0, 0.0], [10.0, 0.0, 0.0]]

        antennas = [Antenna(location.Dunedin, i) for i in antenna_locations]
        src = [
            SimulationSource(
                r=1e9,
                amplitude=1.0,
                azimuth=angle.from_dms(0.0),
                elevation=angle.from_dms(20.0),
                sample_duration=sample_duration,
            ),
            SimulationSource(
                r=1e9,
                amplitude=0.5,
                azimuth=angle.from_dms(0.0),
                elevation=angle.from_dms(20.0),
                sample_duration=sample_duration,
            ),
        ]
        radio_sampling_rate = ref_freq * freq_mult
        tb = np.arange(0, sample_duration, 1.0 / radio_sampling_rate)
        ant_models = [antenna_model.GpsPatchAntenna() for _ in range(2)]

        ant = antennas_signal(antennas, ant_models, src, tb)
        self.assertEqual(ant.shape, (len(antenna_locations), len(tb)))
        mu0 = ant[0, :].mean()
        self.assertLess(mu0, 0.01)
        self.assertGreater(mu0, -0.01)

        mu1 = ant[1, :].mean()
        self.assertLess(mu1, 0.01)
        self.assertGreater(mu1, -0.01)



# Copyright (C) Tim Molteno 2008-2019. All rights reserved

import math

from tart.util import utc


class GpsTime:
    SEC_PER_DAY = 86400.0

    def __init__(self, y, m, d, h, minutes, seconds):
        if y < 1980:
            raise Exception("Year #{y} must >= 1980")

        in_time = utc.utc_datetime(
            year=y, month=m, day=d, hour=h, minute=minutes, second=0
        )
        gps_t = utc.utc_datetime(year=1980, month=1, day=6, hour=0, minute=0, second=0)

        self.m_day_of_year = in_time.timetuple().tm_yday

        gps_sec = (in_time - gps_t).total_seconds() + seconds + self.utc_offset(in_time)
        gps_days = int(math.floor(gps_sec / GpsTime.SEC_PER_DAY))
        gps_weeks = gps_days / 7.0

        self.m_week = int(math.floor(gps_weeks))
        self.sec_of_week = gps_sec - self.m_week * 7.0 * GpsTime.SEC_PER_DAY

    @classmethod
    def from_time(self, utc_date):
        return GpsTime(
            utc_date.year,
            utc_date.month,
            utc_date.day,
            utc_date.hour,
            utc_date.minute,
            utc_date.second,
        )

    # I don't think that the UTC offset works correctly.
    @classmethod
    def utc_offset(self, t):
        if t >= utc.utc_datetime(2017, 1, 1):
            return 18
        if t >= utc.utc_datetime(2015, 7, 1):
            return 17
        if t >= utc.utc_datetime(2012, 7, 1):
            return 16
        if t >= utc.utc_datetime(2009, 1, 1):
            return 15
        if t >= utc.utc_datetime(2006, 1, 1):
            return 14
        if t >= utc.utc_datetime(1999, 1, 1):
            return 13
        if t >= utc.utc_datetime(1997, 7, 1):
            return 12
        if t >= utc.utc_datetime(1996, 1, 1):
            return 11
        if t >= utc.utc_datetime(1994, 7, 1):
            return 10
        if t >= utc.utc_datetime(1993, 7, 1):
            return 9
        if t >= utc.utc_datetime(1992, 7, 1):
            return 8
        if t >= utc.utc_datetime(1991, 1, 1):
            return 7
        if t >= utc.utc_datetime(1990, 1, 1):
            return 6
        if t >= utc.utc_datetime(1988, 1, 1):
            return 5
        if t >= utc.utc_datetime(1985, 7, 1):
            return 4
        if t >= utc.utc_datetime(1983, 7, 1):
            return 3
        if t >= utc.utc_datetime(1982, 7, 1):
            return 2
        if t >= utc.utc_datetime(1981, 7, 1):
            return 1
        if t >= utc.utc_datetime(1980, 1, 1):
            return 0
        return False

    def __repr__(self):
        return "W:%d_SOW:%d (doy %d)" % (
            self.m_week,
            self.sec_of_week,
            self.m_day_of_year,
        )

    def diff(self, b):
        w = (self.m_week - b.week()) * 7 * GpsTime.SEC_PER_DAY
        s = self.sec_of_week - b.sow()
        return w + s

    def week(self):
        return self.m_week

    def sow(self):
        return self.sec_of_week

    def day_of_year(self):
        return self.m_day_of_year

    def day_of_week(self):
        dow = self.sec_of_week / GpsTime.SEC_PER_DAY
        return int(math.floor(dow))



import unittest

import numpy as np

from tart.imaging import antenna_model, correlator
from tart.operation import observation, settings
from tart.simulation import antennas, radio, simulation_source
from tart.util import angle, utc

TEST_CONFIG = "./tart/test/test_telescope_config.json"
TEST_ANTENNA_POSITIONS = "./tart/test/test_calibrated_antenna_positions.json"


def relative_diff(l1, l2):
    return (l1 - l2) / np.sqrt(np.power(l1, 2) + np.power(l2, 2))


class TestSimulatedVisibility(unittest.TestCase):
    def setUp(self):
        self.utc_date = utc.now()
        self.config = settings.from_file(TEST_CONFIG)
        self.config.load_antenna_positions(
            cal_ant_positions_file=TEST_ANTENNA_POSITIONS
        )

        self.rad = radio.Max2769B(
            noise_level=[0.0 for _ in range(self.config.get_num_antenna())]
        )
        self.sources = [
            simulation_source.SimulationSource(
                r=1e9,
                amplitude=1.0,
                azimuth=angle.from_dms(2.0),
                elevation=angle.from_dms(50.0),
                sample_duration=self.rad.sample_duration,
            )
        ]
        self.ants = [
            antennas.Antenna(self.config.get_loc(), pos)
            for pos in self.config.get_antenna_positions()
        ]
        self.ant_models = [
            antenna_model.GpsPatchAntenna()
            for i in range(self.config.get_num_antenna())
        ]
        self.cor = correlator.Correlator()
        self.timebase = np.arange(
            0, self.rad.sample_duration, 1.0 / self.rad.sampling_rate
        )

        ant_sigs_full = antennas.antennas_signal(
            self.ants, self.ant_models, self.sources, self.timebase
        )
        obs = self.rad.get_full_obs(
            ant_sigs_full, self.utc_date, self.config, self.timebase
        )
        vis = self.cor.correlate(obs)
        self.full_vis = np.array(vis.v)

        ant_sigs_simp = antennas.antennas_simplified_signal(
            self.ants,
            self.ant_models,
            self.sources,
            self.rad.baseband_timebase,
            self.rad.int_freq,
        )
        obs = self.rad.get_simplified_obs(
            ant_sigs_simp, self.utc_date, config=self.config
        )
        vis2 = self.cor.correlate(obs)
        self.sim_vis = np.array(vis2.v)

#        plt.figure(figsize=(4, 4))
#        plt.scatter(self.full_vis.real, self.full_vis.imag, label="full")
#        plt.scatter(self.sim_vis.real, self.sim_vis.imag, color="red", label="simpl")
#        plt.legend()
#        plt.show()

    # def test_full_vis(self):
    #   self.assertGreater(self.full_vis.max(), 0.8)
    #   self.assertLess(self.full_vis.min(), -0.8)

    # def test_simp_vis(self):
    #   self.assertGreater(self.sim_vis.max(), 0.8)
    #   self.assertLess(self.sim_vis.min(), -0.8)

    def test_compare_vis(self):
        d = self.full_vis - self.sim_vis
        print(d)
        print("relative diffs", relative_diff(self.full_vis, self.sim_vis))
        self.assertLess(abs(d.max()), 0.3)
        self.assertLess(d.std(), 0.1)


class TestTelescope(unittest.TestCase):
    def setUp(self):

        # rad = radio.Max2769B(noise_level=0.)
        utc_date = utc.now()
        self.config = settings.from_file(TEST_CONFIG)
        self.config.load_antenna_positions(
            cal_ant_positions_file=TEST_ANTENNA_POSITIONS
        )

        rad = radio.Max2769B(
            noise_level=[0.0 for _ in range(self.config.get_num_antenna())]
        )
        sources = [
            simulation_source.SimulationSource(
                r=1e9,
                amplitude=1.0,
                azimuth=angle.from_dms(2.0),
                elevation=angle.from_dms(50.0),
                sample_duration=rad.sample_duration,
            )
        ]
        ant_models = [
            antenna_model.GpsPatchAntenna()
            for i in range(self.config.get_num_antenna())
        ]
        ants = [
            antennas.Antenna(self.config.get_loc(), pos)
            for pos in self.config.get_antenna_positions()
        ]
        self.timebase = np.arange(0, rad.sample_duration, 1.0 / rad.sampling_rate)

        ant_sigs = antennas.antennas_signal(ants, ant_models, sources, self.timebase)
        self.obs = rad.get_full_obs(ant_sigs, utc_date, self.config, self.timebase)

    def test_data_limits(self):
        for i in range(1, self.obs.config.get_num_antenna()):
            data = self.obs.get_antenna(i)

            self.assertGreater(data.max(), 0.99)
            self.assertLess(data.min(), -0.99)

            self.assertGreater(data.mean(), -0.1)
            self.assertLess(data.mean(), 0.1)

    def test_settings(self):
        self.assertEqual(self.config.get_num_antenna(), 24)

    def test_simulated_observation(self):
        fname = "test_observation.pkl"
        self.obs.save(fname)
        nobs = observation.Observation_Load(fname)
        self.assertEqual(self.obs.timestamp, nobs.timestamp)
        self.assertTrue((self.obs.get_antenna(1) == nobs.get_antenna(1)).all())



import unittest

import numpy as np

from tart.util.compare_phases import compare_phases, unwrap


class TestComparePhases(unittest.TestCase):
    def test_unwrap(self):
        torad = np.pi/180.
        ret = unwrap(0)
        self.assertAlmostEqual(ret, 0*torad)
        ret = unwrap(180*torad)
        self.assertAlmostEqual(ret, 180*torad)
        ret = unwrap(3*np.pi)
        self.assertAlmostEqual(ret, np.pi)
        ret = unwrap(-3*np.pi)
        self.assertAlmostEqual(ret, np.pi)
        ret = unwrap(-9*np.pi)
        self.assertAlmostEqual(ret, np.pi)

    def test_compare_phases(self):
        torad = np.pi/180.
        ret = compare_phases(350*torad, 0)
        self.assertAlmostEqual(ret, 10*torad)
        ret = compare_phases(10*torad, 0)
        self.assertAlmostEqual(ret, 10*torad)
        ret = compare_phases(350*torad, -5*torad)
        self.assertAlmostEqual(ret, 5*torad)
        ret = compare_phases(-5*torad, 350*torad)
        self.assertAlmostEqual(ret, 5*torad)
        ret = compare_phases(10*torad, 20*torad)
        self.assertAlmostEqual(ret, 10*torad)
        ret = compare_phases(10*torad, 350*torad)
        self.assertAlmostEqual(ret, 20*torad)
        ret = compare_phases(-10*torad, -30*torad)
        self.assertAlmostEqual(ret, 20*torad)
        ret = compare_phases(-900.*torad, -0.*torad)
        self.assertAlmostEqual(ret, 180*torad)



## Utility functions for imaging
import numpy as np

from tart.imaging import synthesis
from tart.operation import settings


def deg_to_pix(num_bins, deg):
    pix_per_degree = num_bins / 180.0
    d = deg*pix_per_degree
    return d

        # x_px = int(np.round(self.l * n2 + n2))
        # y_px = num_bins - int(np.round(self.m * n2 + n2))


def get_l_index(l, image_size):
    image_radius = image_size / 2
    return int(np.round(l * image_radius + image_radius))


def get_m_index(m, image_size):
    image_radius = image_size / 2
    return image_size - int(np.round(m * image_radius + image_radius))



def get_baseline_indices(num_ant):
    bl_indices = []
    for i in range(num_ant-1):
        for j in range(i + 1, num_ant):
            bl_indices.append([i, j])
    return bl_indices


def ant_pos_to_uv(ant_pos, i, j):
    return ant_pos[j] - ant_pos[i]


def apply_complex_gains(v_complex, gains_complex, i, j):
    return v_complex * gains_complex[i] * np.conj(gains_complex[j])


def ifft_imaging(uv_plane, module=np.fft):
    return module.fft.fftshift(
        module.fft.ifft2(
            module.fft.ifftshift(uv_plane)
            )
        )


def uv_index(u, v, num_bins, uv_max):
    ''' A little function to produce the index into the u-v array
        for a given value (u, measured in wavelengths)
    '''
    middle = num_bins // 2

    u_pix = middle + (u / uv_max)*(num_bins/2)
    v_pix = middle + (v / uv_max)*(num_bins/2)

    return int(u_pix), int(v_pix)


def grid_visibility(uv_plane, v_complex, baselines):
    num_bins = uv_plane.shape[0]
    uv_max = num_bins / 4   # (1.2 * np.pi)

    n_vis = len(v_complex)
    for i in range(n_vis):
        v = v_complex[i]
        uu, vv, ww = baselines[i]

        u_idx, v_idx = uv_index(uu, vv, num_bins, uv_max)
        uv_plane[u_idx, v_idx] += v

        # Place the conjugate visibility at -uu, -vv
        u_idx, v_idx = uv_index(--uu, --vv, num_bins, uv_max)
        uv_plane[u_idx, v_idx] += np.conj(v)

    return uv_max


def rotate_vis(rot_degrees, cv, reference_positions):
    '''
        Note. This rotates counter_clockwise
        (antennas in the north move towards the east)
    '''
    conf = cv.vis.config

    new_positions = settings.rotate_location(
        rot_degrees, np.array(reference_positions).T
    )
    conf.set_antenna_positions((np.array(new_positions).T).tolist())


def image_from_calibrated_vis(cv, nw, num_bin):
    cal_syn = synthesis.Synthesis_Imaging([cv])

    cal_ift, cal_extent = cal_syn.get_ift(nw=nw, num_bin=num_bin)
    # beam = cal_syn.get_beam(nw=nw, num_bin=num_bin, use_kernel=False)
    n_fft = len(cal_ift)
    assert n_fft == num_bin

    bin_width = (max(cal_extent) - min(cal_extent)) / float(n_fft)

    return cal_ift, cal_extent, n_fft, bin_width



import numpy as np

# Displacement is the (+/-) displacement of an antenna from the origin
# Angle is in degrees. Clockwise angles from vertical +ve, anticlockwise -ve.
# A +ve delay indicates antenna leads center, -ve that it lags.
# Variable "sample_delay" is in seconds.


def sample_delay(displacement, angle):
    sample_delay = displacement * np.sin(np.radians(angle)) / 2.99793e8
    return sample_delay


if __name__ == "__main__":

    displacement = -5
    angle = 45
    delay = sample_delay(displacement, angle)
    print(delay)



import numpy as np

from tart.util import angle


def unit_vector(vector):
    """ Returns the unit vector of the vector.    """
    return vector / np.linalg.norm(vector)

def angle_between(v1, v2):
    """ Returns the angle in radians between vectors 'v1' and 'v2'::

        >>> angle_between((1, 0, 0), (0, 1, 0))
        1.5707963267948966
        >>> angle_between((1, 0, 0), (1, 0, 0))
        0.0
        >>> angle_between((1, 0, 0), (-1, 0, 0))
        3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    theta = np.arccos(np.dot(v1_u, v2_u))
    if np.isnan(theta):
        if (v1_u == v2_u).all():
            return angle.from_rad(0.0)
        else:
            return angle.from_rad(np.pi)
    return angle.from_rad(theta)



#
# Helper class for dealing with angles (and to avoid getting tangled between degrees and radians.
#
# Copyright (c) Tim Molteno 2013. tim@elec.ac.nz
#
import math

import numpy as np

from tart.util import constants


def deg_to_rad(x):
    return x * math.pi / 180.0


def wrap_2pi(x):
    r = math.fmod(x, constants.TWO_PI)
    if (r < 0.0):
        r += constants.TWO_PI
    return r


def wrap_360(x):
    r = math.fmod(x, 360.0)
    if (r < 0.0):
        r += 360.0
    return r


def wrap_180(x):
    r = math.fmod(x, 360.0)
    if (r > 180.0):
        r -= 360.0
    return r


# from functools import total_ordering
# @total_ordering
class Angle:

    def __init__(self, deg=0, minute=0, sec=0):
        dec_deg = deg + float(minute)/60.0 + float(sec) / 3600.0
        self.rad = deg_to_rad(dec_deg)

    def __repr__(self):
        return str(self.to_degrees())

    def to_elevation(self):
         deg = self.to_degrees()
         if (deg > 90.0):
             raise "Elevation out of RANGE %f" % deg
         if (deg < -90.0):
             raise "Elevation out of RANGE %f" % deg

         return from_dms(deg)

    def to_declination(self):
        return self.to_elevation()

    def to_ra(self):
        return from_rad(wrap_2pi(self.rad))

    def to_rad(self):
        return self.rad

    def to_degrees(self):
        return self.rad * (360.0 / constants.TWO_PI)

    def to_hours(self):
        return wrap_2pi(self.rad) * (24.0 / constants.TWO_PI)

    def to_dms(self):
        dd = self.to_degrees()
        mnt,sec = divmod(dd*3600,60) ## 60 seconds in a
        deg,mnt = divmod(mnt,60)
        return deg,mnt,sec

    def to_hms(self):
        dd = self.to_degrees()

        hour,rem = divmod(dd, 15)   # 15 Degrees in an hour
        mnt,sec = divmod(rem*3600/15,60)  # 60 Minutes in an hour
        return hour,mnt,sec

    def sin(self):
        return math.sin(self.rad)

    def cos(self):
        return math.cos(self.rad)

    def tan(self):
        return math.tan(self.rad)

    def __abs__(self):
        return from_rad(abs(self.rad))

    def __add__(self, x):
        return from_rad(self.rad + x.rad)

    def __radd__(self, x):
        return self.__add__(x)

    def __sub__(self, x):
        return from_rad(self.rad - x.rad)

    def __rsub__(self, x):
        return self.__sub__(x)

    def __eq__(self, x):
        return (wrap_2pi(self.rad) == wrap_2pi(x.rad))

    def __lt__(self, x):
        return self.rad < x.rad # TODO should we wrap both?

    def __neg__(self):
        return from_rad(-self.rad)

def from_hours(hour, minute=0.0, sec=0.0):
    hr = hour % 24.0
    return Angle(deg=hr*15, minute=minute*15, sec=sec*15)

def from_dms(deg=0.0, minute=0.0, sec=0.0):
    return Angle(deg, minute, sec)

def from_rad(rad):
    a = Angle(0.,0.,0.)
    a.rad = rad
    return a

def atan(x):
    return from_rad(np.arctan(x))

def atan2(x,y):
    return from_rad(np.arctan2(x,y))

def asin(x):
    return from_rad(np.arcsin(x))

def acos(x):
    return from_rad(np.arccos(x))

def add_rad(a, x):
    return from_rad(a.rad + x)



import numpy as np

from tart.imaging import visibility
from tart.util import angle

# import pyfftw
# from pyfftw.interfaces.scipy_fftpack import hilbert as fftw_hilbert
# from scipy.fftpack import hilbert as fftw_hilbert
# from tart.util.hilbert import hilbert_fftw as fftw_hilbert


def van_vleck_correction(R):
    '''
        Correct the correlation for two-level quantization.
        https://arxiv.org/abs/1608.04367

        - R the output of the digital correlator
    '''
    return np.sin((np.pi / 2.0) * R)


def combine_real_imag(v_real, v_imag):
    return v_real - 1j * v_imag


class Correlator:
    def __init__(self, van_vleck_corr=True):
        self.vv = van_vleck_corr

    def correlate(self, obs, debug=False, mode="roll"):
        visibilities, baselines = self.compute_complex_vis(obs, debug=debug, mode=mode)
        vis = visibility.Visibility.from_config(obs.config, obs.timestamp)
        vis.set_visibilities(visibilities, baselines)
        return vis

    def compute_complex_vis(self, obs, debug=False, mode="roll"):
        """Return an array of baselines and visibilities from this observation"""
        v = []
        baselines = []
        data = []
        data_hilb = []
        num_antenna = obs.config.get_num_antenna()
        for i in range(num_antenna):
            ant_i = obs.get_antenna(i)
            mean_i = np.mean(ant_i)
            data.append(ant_i - mean_i)

        if "fftw" in mode:
            from pyfftw.interfaces.scipy_fftpack import hilbert as fftw_hilbert

        for i, d in enumerate(data):
            if mode == "roll":
                data_hilb.append(np.roll(d, 1))
            elif mode == "fftw_hilbert":
                h_i = -fftw_hilbert(d)
                data_hilb.append(h_i)
            elif mode == "fftw_hilbert_sign":
                h_i = -np.sign(fftw_hilbert(d))
                data_hilb.append(h_i)

        for i in range(num_antenna):
            for j in range(i + 1, num_antenna):
                v.append(self.V(data[i], data[j], data_hilb[j]))
                baselines.append([i, j])
        return v, baselines

    def V(self, x, y, yhilb):
        v_real = np.dot(x, y) * 1.0 / float(len(x))
        v_imag = np.dot(x, yhilb) * 1.0 / float(len(x))
        if self.vv:
            v_real = van_vleck_correction(v_real)
            v_imag = van_vleck_correction(v_imag)
        return combine_real_imag(v_real, v_imag)

    # def correlate_roll(self, obs, debug=False):
    # v = []
    # baselines = []
    # num_ant = obs.config.get_num_antenna()
    # data = obs.data # use obs.data[i] instead of get_antenna to keep mem usage low
    # means = obs.get_means()
    # cos_str = []
    # sin_str = []
    # for a in data:
    # cos_str.append(a[1:].tostring())
    # sin_str.append(a[:-1].tostring())
    # n = len(data[0][1:])
    # for i in range(0, num_ant):
    # for j in range(i+1, num_ant):
    # if debug:
    # progress = len(baselines)
    # if (progress%10==0):
    # print(progress)
    ##cos_i = data[i][1:]
    ##cos_j = data[j][1:]
    ##sin_j = data[j][:-1]
    ##v_real =    -means[i]*means[j] + corr_b(cos_i, cos_j, n)
    ##v_imag =    -means[i]*means[j] + corr_b(cos_i, sin_j, n)
    # v_real =    -means[i]*means[j] + corr_b_cpp(cos_str[i], cos_str[j], n)
    # v_imag =    -means[i]*means[j] + corr_b_cpp(cos_str[i], sin_str[j], n)

    # if self.vv:
    # v_real = van_vleck_correction(v_real)
    # v_imag = van_vleck_correction(v_imag)
    # v_com = combine_real_imag(v_real,v_imag)
    # v.append(v_com)
    # baselines.append([i,j])
    # vis = visibility.Visibility.from_config(obs.config, obs.timestamp)
    # vis.set_visibilities(v, baselines)
    # return vis


def corr_b(x, y, n):
    # num_not_same = (x ^ y).sum()
    assert n == len(y)
    num_not_same = np.count_nonzero(x ^ y)  # alot faster than sum.
    ret = 1 - 2 * num_not_same / float(n)
    return ret


def corr_b_pat(x, y):
    n = len(x)
    num_same = (x ^ (1 - y)).sum()
    ret = 2 * num_same / float(n) - 1
    return ret


# from scipy import weave

# code = """
# PyObject* res = PyString_FromStringAndSize(NULL, real_size);

# const ssize_t tail = (ssize_t)PyString_AS_STRING(res) % ALIGNMENT;
# const ssize_t head = (ALIGNMENT - tail) % ALIGNMENT;

# memxor((const char*)a, (const char*)b, PyString_AS_STRING(res), head);

# const __m128i* pa = (__m128i*)((char*)a + head);
# const __m128i* pend = (__m128i*)((char*)a + real_size - tail);
# const __m128i* pb = (__m128i*)((char*)b + head);
# __m128i xmm1, xmm2;
# __m128i* pc = (__m128i*)(PyString_AS_STRING(res) + head);
# while (pa < pend) {
# xmm1 = _mm_loadu_si128(pa);
# xmm2 = _mm_loadu_si128(pb);
# _mm_stream_si128(pc, _mm_xor_si128(xmm1, xmm2));
# ++pa;
# ++pb;
# ++pc;
# }
# memxor((const char*)pa, (const char*)pb, (char*)pc, tail);
# return_val = res;
# Py_DECREF(res);
# """

# support = """
##define ALIGNMENT 16
# static void memxor(const char* in1, const char* in2, char* out, ssize_t n) {
# const char* end = in1 + n;
# while (in1 < end) {
# *out = *in1 ^ *in2;
# ++in1;
# ++in2;
# ++out;
# }
# }
# """

# def corr_b_cpp(x, y, n):
# a = np.frombuffer(x, dtype=np.bool)
# b = np.frombuffer(y, dtype=np.bool)
# real_size = len(a)
# out = weave.inline(code, ["a", "b", "real_size"],
# headers = ['"emmintrin.h"'],
# support_code = support)
# num_not_same = np.count_nonzero(np.frombuffer(out,dtype=np.bool))
# ret = 1 - 2*num_not_same/float(n)
# return ret


from tart.util import constants


class FxCorrelator(Correlator):
    """A FX correlator.
    We need an FX correlator because the bandwidth is sufficient that it washes out the fringes away
    from the phase center. The distance we can get away from the phase center is given by

    q = arcsin(c / B D)

    where B is the baseline length, D is the bandwidth, c is the speed of light.

    where

           c = 3e8 (m s^-1)
           B = baseline (m)
           D = bw (s^-1)
    """

    def __init__(self, bandwidth, linewidth):
        self.bandwidth = bandwidth
        n_log2 = int(np.log(bandwidth / linewidth) / np.log(2.0) + 0.5)
        self.n_freq = 2 ** n_log2
        self.linewidth = bandwidth / self.n_freq

    def get_linewidth(self):
        return self.linewidth

    def correlate(self, obs):
        data = self.convert_to_baseband(obs)
        vis = visibility.Visibility.from_obs(obs, angle.from_dms(90), angle.from_dms(0))
        visibilities, baselines = self.compute_complex_vis(obs)
        vis.set_visibilities(visibilities, baselines)
        return vis

    """TODO this is not working yet.
         this should return the maximum field of view possible before fringes get washed
         out by the bandwidth of the signal."""

    def angular_resolution(self, baseline):
        x = constants.V_LIGHT / (self.linewidth * baseline)
        if x >= 1.0:
            x = 1.0
        return angle.asin(x)



"""
Solar Radio Source

Calculate the position in horizontal coordinates for the sun.
"""
#
# Copyright (c) Tim Molteno 2013. tim@elec.ac.nz
#
# import datetime
# import math
# from tart.util import angle


import unittest

import numpy as np

from tart.imaging.elaz import ElAz


class TestElaz(unittest.TestCase):

    def test_elaz_lmn_vertical(self):
        elaz = ElAz(90, 0)
        elaz2 = ElAz(90, 90)

        self.assertAlmostEqual(elaz.l, 0)
        self.assertAlmostEqual(elaz.m, 0)
        self.assertAlmostEqual(elaz2.l, 0)
        self.assertAlmostEqual(elaz2.m, 0)

    def test_elaz_lmn_horizontal(self):
        elaz = ElAz(0, 0)    # Due north (straight up) 0, 1
        elaz2 = ElAz(0, 90)  # Due west (-1,0)

        self.assertAlmostEqual(elaz.l, 0)
        self.assertAlmostEqual(elaz.m, 1)
        self.assertAlmostEqual(elaz2.l, -1)
        self.assertAlmostEqual(elaz2.m, 0)

    def test_elaz_px_window(self):
        elaz = ElAz(90, 0)

        x_min, x_max, y_min, y_max, area = elaz.get_px_window(num_bins=128, window_deg=6)
        self.assertAlmostEqual(x_min, 59)
        self.assertAlmostEqual(x_max, 69)
        x_min, x_max, y_min, y_max, area = elaz.get_px_window(num_bins=128, window_deg=3)
        self.assertAlmostEqual(x_min, 61)
        self.assertAlmostEqual(x_max, 67)

        self.assertAlmostEqual(y_min, 61)
        self.assertAlmostEqual(y_max, 67)

    def test_elaz_px_vertical(self):

        elaz = ElAz(90, 0)
        elaz2 = ElAz(90, 90)

        x, y = elaz.get_px(num_bins=128)
        self.assertAlmostEqual(x, 64)
        self.assertAlmostEqual(y, 64)

        x, y = elaz2.get_px(num_bins=128)
        self.assertAlmostEqual(x, 64)
        self.assertAlmostEqual(y, 64)

    def test_elaz_px_horizontal(self):

        elaz = ElAz(0, 0)    # Due north (straight up, center) 0, 1
        elaz2 = ElAz(0, 90)  # Due west (-1,0)

        x, y = elaz.get_px(num_bins=128)
        self.assertAlmostEqual(x, 64)
        self.assertAlmostEqual(y, 0)

        elaz_r = ElAz.from_pixels(x, y, num_bins=128)
        self.assertEqual(elaz_r.el_r, 0)
        self.assertEqual(elaz_r.az_r, 0)

        x, y = elaz2.get_px(num_bins=128)
        self.assertAlmostEqual(x, 0)
        self.assertAlmostEqual(y, 64)

        elaz_r = ElAz.from_pixels(x, y, num_bins=128)
        self.assertEqual(elaz_r.el_r, 0)
        self.assertEqual(elaz_r.az_r, np.radians(90))

    def test_round_trip(self):
        num_bins = 512
        for i in range(100):
            elaz = ElAz(np.random.uniform(1, 89),
                        np.random.uniform(0, 360))
            print(f"random elaz: {elaz}")
            l, m = elaz.get_lm()
            print(f"    l={l}, m={m}")
            x, y = elaz.get_px(num_bins)
            print(f"    x={x}, y={y}")
            # Now construct elaz from x,y

            elaz2 = ElAz.from_pixels(x, y, num_bins)
            print(f"reconstructed elaz: {elaz2}")

            self.assertAlmostEqual(elaz.el_r, elaz2.el_r, 1)
            if elaz.el_r < np.radians(60):
                self.assertAlmostEqual(np.sin(elaz.az_r), np.sin(elaz2.az_r), 1)
